import "reflect-metadata";
import { Container } from "inversify";
import { API } from "lambda-api";
import { Server } from "./api/Server";
import { AppConfig } from "./model/AppConfig";
import { ApiRequest } from "./model/ApiRequest";
import { ApiResponse } from "./model/ApiResponse";
/**
 * Application base class which combines the `Server`, `Container`(`InversifyJS`)
 * and `AppConfig` classes to create a decorator driven API with typescript
 * middleware and dependency injection. It uses the `lambda-api` package as the
 * underlying HTTP API framework.
 *
 * AWS Lambda requests are handled by the `run` method, which will return a response
 * compatible with either API Gateway or an ALB.
 *
 * Extending this class will allow creating an app implementation for runtimes,
 * AWS Lambda or local web server for example.
 */
export declare abstract class ApiApp {
    protected readonly controllersPath: string;
    protected appConfig: AppConfig;
    protected appContainer: Container;
    protected readonly apiServer: Server;
    readonly middlewareRegistry: import("./typescript-lambda-api").MiddlewareRegistry;
    /**
     * Create a new app.
     *
     * @param controllersPath Path to a directory containing `js` files containing that declare controllers.
     * @param appConfig (Optional) Application config to pass to `lambda-api`.
     * @param appContainer (Optional) `InversifyJS` IOC `Container` instance which can
     *                     build controllers and error interceptors.
     */
    constructor(controllersPath: string, appConfig?: AppConfig, appContainer?: Container);
    /**
     * Configure the `InversifyJS` IOC `Container` instance.
     *
     * @param configureBlock Function that takes a `Container` instance as a parameter.
     */
    configureApp(configureBlock: (this: void, container: Container) => void): void;
    /**
     * Configure the `API` instance from the `lambda-api` package.
     *
     * @param handler Function that takes an `API` instance as a parameter.
     */
    configureApi(configureBlock: (this: void, api: API) => void): void;
    /**
     * Run using the passed event and context, ultimately should call the
     * `processEvent` method on the `apiServer` instance.
     *
     * @param request API Gateway or ALB request.
     * @param context Request context.
     * @returns The response.
     */
    abstract run(event: ApiRequest, context: any): Promise<ApiResponse>;
    /**
     * Initialise all controllers and endpoints declared using decorators.
     */
    initialiseControllers(): Promise<void>;
}
