"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
const inversify_1 = require("inversify");
const Server_1 = require("./api/Server");
const AppConfig_1 = require("./model/AppConfig");
/**
 * Application base class which combines the `Server`, `Container`(`InversifyJS`)
 * and `AppConfig` classes to create a decorator driven API with typescript
 * middleware and dependency injection. It uses the `lambda-api` package as the
 * underlying HTTP API framework.
 *
 * AWS Lambda requests are handled by the `run` method, which will return a response
 * compatible with either API Gateway or an ALB.
 *
 * Extending this class will allow creating an app implementation for runtimes,
 * AWS Lambda or local web server for example.
 */
class ApiApp {
    /**
     * Create a new app.
     *
     * @param controllersPath Path to a directory containing `js` files containing that declare controllers.
     * @param appConfig (Optional) Application config to pass to `lambda-api`.
     * @param appContainer (Optional) `InversifyJS` IOC `Container` instance which can
     *                     build controllers and error interceptors.
     */
    constructor(controllersPath, appConfig = new AppConfig_1.AppConfig(), appContainer = new inversify_1.Container({ autoBindInjectable: true })) {
        this.controllersPath = controllersPath;
        this.appConfig = appConfig;
        this.appContainer = appContainer;
        this.apiServer = new Server_1.Server(appContainer, appConfig);
    }
    get middlewareRegistry() {
        return this.apiServer.middlewareRegistry;
    }
    /**
     * Configure the `InversifyJS` IOC `Container` instance.
     *
     * @param configureBlock Function that takes a `Container` instance as a parameter.
     */
    configureApp(configureBlock) {
        configureBlock(this.appContainer);
    }
    /**
     * Configure the `API` instance from the `lambda-api` package.
     *
     * @param handler Function that takes an `API` instance as a parameter.
     */
    configureApi(configureBlock) {
        this.apiServer.configure(configureBlock);
    }
    /**
     * Initialise all controllers and endpoints declared using decorators.
     */
    async initialiseControllers() {
        await this.apiServer.discoverAndBuildRoutes(this.controllersPath);
    }
}
exports.ApiApp = ApiApp;
//# sourceMappingURL=ApiApp.js.map