"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const AuthResult_1 = require("../model/security/AuthResult");
class Endpoint {
    constructor(endpointInfo, controllerFactory, errorInteceptorFactory, middlewareRegistry) {
        this.endpointInfo = endpointInfo;
        this.controllerFactory = controllerFactory;
        this.errorInteceptorFactory = errorInteceptorFactory;
        this.middlewareRegistry = middlewareRegistry;
    }
    register(api) {
        let registerMethod = this.mapHttpMethodToCall(api, this.endpointInfo.httpMethod);
        registerMethod(this.endpointInfo.fullPath, async (req, res) => await this.invoke(req, res));
    }
    // entry point for lambda-api request engine
    async invoke(request, response) {
        let principal = await this.authenticateAndAuthorizeRequest(request, response);
        if (this.responseSent(response)) {
            // 401 or 403 response was sent
            return;
        }
        let controller = this.buildControllerInstance(request, response);
        this.setResponseContentType(response);
        let endpointResponse = await this.invokeControllerMethod(controller, request, response, principal);
        if (!this.responseSent(response, endpointResponse)) {
            throw new Error("no response was sent by or value returned from endpoint method, " +
                `path: ${this.endpointInfo.path} | endpoint: ${this.endpointInfo.name}`);
        }
        return endpointResponse;
    }
    async authenticateAndAuthorizeRequest(request, response) {
        if (this.endpointInfo.authenticationDisabled) {
            return;
        }
        let authResult = await this.authenticateRequest(request);
        if (!authResult.authenticated) {
            this.sendStatusCodeResponse(401, response);
        }
        if (!await this.authorizeRequest(authResult.principal)) {
            this.sendStatusCodeResponse(403, response);
        }
        return authResult.principal;
    }
    async authenticateRequest(request) {
        let authResult = new AuthResult_1.AuthResult();
        if (this.middlewareRegistry.authFilters.length < 1) {
            authResult.authenticated = true;
            return authResult;
        }
        for (let filter of this.middlewareRegistry.authFilters) {
            let authData = await filter.extractAuthData(request);
            if (authData) {
                // authData extraction was either aborted or failed
                authResult.principal = await filter.authenticate(authData);
            }
            authResult.authenticated =
                (authResult.principal !== null && authResult.principal !== undefined);
            if (authResult.authenticated) {
                // return after finding a filter that authenticates the user
                return authResult;
            }
        }
        return authResult;
    }
    sendStatusCodeResponse(statusCode, response) {
        response.status(statusCode)
            .removeHeader("content-type")
            .header("content-type", "text/plain")
            .send("");
    }
    async authorizeRequest(principal) {
        let roles = this.endpointInfo.roles;
        if (!roles || roles.length < 1) {
            return true;
        }
        if (this.middlewareRegistry.authorizers.length < 1) {
            throw new Error("Role restrictions were declared but no authorizer was registered in the middleware registry, " +
                `path: ${this.endpointInfo.path} | endpoint: ${this.endpointInfo.name}`);
        }
        for (let authorizer of this.middlewareRegistry.authorizers) {
            // endpoint roles, if defined, override controller roles
            for (let role of roles) {
                if (await authorizer.authorize(principal, role)) {
                    return true;
                }
            }
        }
        return false;
    }
    responseSent(response, endpointResponse) {
        let rawRes = response;
        return endpointResponse || (rawRes && rawRes._state === "done");
    }
    buildControllerInstance(request, response) {
        if (!this.endpointInfo.controller) {
            return this.buildDynamicControllerInstance(request, response);
        }
        let controller = this.controllerFactory(this.endpointInfo.controller.classConstructor);
        if (typeof (controller.setRequest) === "function") {
            controller.setRequest(request);
        }
        if (typeof (controller.setResponse) === "function") {
            controller.setResponse(response);
        }
        return controller;
    }
    /**
     * If an endpoint is function only (has no controller bound to it), we
     * build a dynamic controller which simulates a controller instance.
     */
    buildDynamicControllerInstance(request, response) {
        let dynamicController = {
            request,
            response
        };
        dynamicController[this.endpointInfo.methodName] = this.endpointInfo.method;
        return dynamicController;
    }
    setResponseContentType(response) {
        let produces = this.endpointInfo.responseContentType;
        if (produces) {
            response.removeHeader("Content-Type")
                .header("Content-Type", produces);
        }
    }
    mapHttpMethodToCall(api, method) {
        if (method === "GET") {
            return api.get;
        }
        else if (method === "POST") {
            return api.post;
        }
        else if (method === "PUT") {
            return api.put;
        }
        else if (method === "PATCH") {
            return api.patch;
        }
        else if (method === "DELETE") {
            return api.delete;
        }
        throw new Error(`Unrecognised HTTP method ${method}`);
    }
    async invokeControllerMethod(controller, request, response, principal) {
        let method = controller[this.endpointInfo.methodName];
        let parameters = this.buildEndpointParameters(request, response, principal);
        try {
            return await method.apply(controller, parameters);
        }
        catch (ex) {
            let errorInterceptor = this.getMatchingErrorInterceptor();
            if (errorInterceptor) {
                let interceptorResponse = await errorInterceptor.intercept({
                    endpointController: controller,
                    endpointMethod: method,
                    endpointMethodParameters: parameters,
                    error: ex,
                    request,
                    response
                });
                if (this.responseSent(response, interceptorResponse)) {
                    return interceptorResponse;
                }
            }
            throw ex;
        }
    }
    buildEndpointParameters(request, response, principal) {
        return this.endpointInfo
            .parameterExtractors
            .map(pe => pe.extract(request, response, principal));
    }
    getMatchingErrorInterceptor() {
        let decoratorInterceptor;
        if (this.endpointInfo.endpointErrorInterceptor) {
            decoratorInterceptor = this.errorInteceptorFactory(this.endpointInfo.endpointErrorInterceptor);
        }
        if (decoratorInterceptor) {
            decoratorInterceptor.controllerTarget = this.endpointInfo.getControllerPropOrDefault(c => c.name);
            decoratorInterceptor.endpointTarget = this.endpointInfo.name;
            return decoratorInterceptor;
        }
        return this.middlewareRegistry.errorInterceptors
            .find(i => i.shouldIntercept(this.endpointInfo.getControllerPropOrDefault(c => c.name), this.endpointInfo.name));
    }
}
exports.Endpoint = Endpoint;
//# sourceMappingURL=Endpoint.js.map