"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const openapi3_ts_1 = require("openapi3-ts");
const MiddlewareRegistry_1 = require("../MiddlewareRegistry");
const DecoratorRegistry_1 = require("../reflection/DecoratorRegistry");
const BasicAuthFilter_1 = require("../security/BasicAuthFilter");
const AppConfig_1 = require("../../model/AppConfig");
const jsonUtils_1 = require("../../util/jsonUtils");
const timed_1 = require("../../util/timed");
class OpenApiGenerator {
    static buildOpenApiSpec(appConfig, middlewareRegistry) {
        return OpenApiGenerator.generateApiOpenApiSpecBuilder(appConfig, middlewareRegistry).getSpec();
    }
    static async exportOpenApiSpec(format = "json", appConfig, middlewareRegistry) {
        let openApiBuilder = OpenApiGenerator.generateApiOpenApiSpecBuilder(appConfig, middlewareRegistry);
        if (format === "json") {
            return openApiBuilder.getSpecAsJson();
        }
        try {
            const jsyaml = await Promise.resolve().then(() => __importStar(require("js-yaml")));
            return jsyaml.safeDump(openApiBuilder.getSpec());
        }
        catch (ex) {
            // TODO: log warning to check that `js-yaml` needs to be installed for yml specs
            throw ex;
        }
    }
    static generateApiOpenApiSpecBuilder(appConfig, middlewareRegistry) {
        let openApiBuilder = openapi3_ts_1.OpenApiBuilder.create();
        let paths = {};
        let tags = {};
        if (appConfig.name) {
            openApiBuilder.addTitle(appConfig.name);
        }
        if (appConfig.version) {
            openApiBuilder.addVersion(appConfig.version);
        }
        OpenApiGenerator.discoverSecuritySchemes(openApiBuilder, middlewareRegistry);
        OpenApiGenerator.discoverTagsAndPaths(paths, tags);
        // add all discovered tags
        for (let tag in tags) {
            if (!tags.hasOwnProperty(tag)) {
                continue;
            }
            openApiBuilder.addTag(tags[tag]);
        }
        // add all discovered paths
        for (let path in paths) {
            if (!paths.hasOwnProperty(path)) {
                continue;
            }
            openApiBuilder.addPath(path, paths[path]);
        }
        return openApiBuilder;
    }
    static discoverSecuritySchemes(openApiBuilder, middlewareRegistry) {
        for (let authFilter of middlewareRegistry.authFilters) {
            let constructor = authFilter.constructor;
            let authFilterInfo = OpenApiGenerator.AUTH_FILTERS[constructor.name];
            if (!authFilterInfo && (authFilter instanceof BasicAuthFilter_1.BasicAuthFilter)) {
                authFilterInfo = OpenApiGenerator.AUTH_FILTERS[BasicAuthFilter_1.BasicAuthFilter.name];
            }
            if (authFilterInfo) {
                openApiBuilder = openApiBuilder.addSecurityScheme(authFilterInfo.name, authFilterInfo.securitySchemeInfo);
            }
        }
    }
    static discoverTagsAndPaths(paths, tags) {
        for (let endpoint in OpenApiGenerator.ENDPOINTS) {
            if (!OpenApiGenerator.ENDPOINTS.hasOwnProperty(endpoint)) {
                continue;
            }
            let endpointInfo = OpenApiGenerator.ENDPOINTS[endpoint];
            if (endpointInfo.controller) {
                OpenApiGenerator.addTagIfPresent(tags, endpointInfo.controller);
            }
            OpenApiGenerator.addEndpoint(paths, endpointInfo);
        }
    }
    static addTagIfPresent(tags, controller) {
        if (!controller.apiName || tags[controller.apiName]) {
            // tag name not defined or already recorded
            return;
        }
        tags[controller.apiName] = {
            description: controller.apiDescription,
            name: controller.apiName
        };
    }
    static addEndpoint(paths, endpointInfo) {
        let path = endpointInfo.fullPath;
        if (path.length > 1 && path.endsWith("/")) {
            // trim trailing forward slash from path
            path = path.substring(0, path.length - 1);
        }
        let pathInfo = paths[path] || {};
        let endpointMethod = endpointInfo.httpMethod.toLowerCase();
        let endpointOperation = {
            responses: {}
        };
        if (endpointMethod !== "get" && endpointMethod !== "delete") {
            OpenApiGenerator.setRequestInfo(endpointOperation, endpointInfo);
        }
        if (endpointInfo.responseContentType) {
            // user declared response content type
            OpenApiGenerator.setEndpointResponseContentType(endpointOperation, endpointInfo.responseContentType);
        }
        else {
            // default response content type
            OpenApiGenerator.setEndpointResponseContentType(endpointOperation, "application/json" // use lambda-api content-type default
            );
        }
        if (endpointInfo.apiOperationInfo) {
            // user declared endpoint info
            OpenApiGenerator.addEndpointOperationInfo(endpointInfo, endpointOperation);
        }
        OpenApiGenerator.addParametersToEndpoint(endpointOperation, endpointInfo);
        if (endpointInfo.getControllerPropOrDefault(c => c.apiName)) {
            // associate endpoint with controller tag name
            endpointOperation.tags = [endpointInfo.controller.apiName];
        }
        else if (endpointInfo.controller) {
            endpointOperation.tags = [endpointInfo.controller.name.replace(/Controller$/, "")];
        }
        else {
            endpointOperation.tags = [endpointInfo.name];
        }
        if (endpointInfo.noAuth) {
            // no authentication required, exclude from global security constraints
            endpointOperation.security = [];
        }
        pathInfo[endpointMethod] = endpointOperation;
        paths[path] = pathInfo;
    }
    static setRequestInfo(endpointOperation, endpointInfo) {
        let requestInfo = endpointInfo.apiRequestInfo;
        let operationRequestType;
        if (requestInfo) {
            operationRequestType = requestInfo.contentType ||
                endpointInfo.requestContentType ||
                "application/json";
        }
        operationRequestType = operationRequestType ||
            endpointInfo.requestContentType;
        if (!operationRequestType) {
            return;
        }
        operationRequestType = operationRequestType.toLowerCase();
        // user declared request content type
        let mediaTypeObject = OpenApiGenerator.setEndpointRequestContentType(endpointOperation, operationRequestType, requestInfo);
        if (!requestInfo || (!requestInfo.type && !requestInfo.class)) {
            // no type info for request body
            return;
        }
        mediaTypeObject.schema = {};
        if (requestInfo.type) {
            OpenApiGenerator.addPrimitiveToMediaTypeObject(mediaTypeObject, requestInfo);
        }
        else if (requestInfo.class) {
            OpenApiGenerator.addClassToMediaTypeObject(mediaTypeObject, requestInfo, operationRequestType);
        }
        if (requestInfo.example) {
            mediaTypeObject.example = requestInfo.example;
        }
        else if (mediaTypeObject.schema) {
            let schema = mediaTypeObject.schema;
            if (schema && schema.example) {
                mediaTypeObject.example = schema.example;
            }
        }
    }
    static setEndpointRequestContentType(endpointOperation, requestContentType, requestInfo) {
        let requestBody = {
            content: {},
            description: requestInfo ? (requestInfo.description || "") : ""
        };
        let mediaTypeObject = {};
        requestBody.content[requestContentType] = mediaTypeObject;
        endpointOperation.requestBody = requestBody;
        return mediaTypeObject;
    }
    static addEndpointOperationInfo(endpointInfo, endpointOperation) {
        let operationInfo = endpointInfo.apiOperationInfo;
        let responseContentType = endpointInfo.responseContentType || "application/json";
        endpointOperation.summary = operationInfo.name;
        endpointOperation.description = operationInfo.description;
        if (operationInfo.request) {
            OpenApiGenerator.setRequestInfo(endpointOperation, endpointInfo);
        }
        for (let statusCode in operationInfo.responses) {
            if (!operationInfo.responses.hasOwnProperty(statusCode)) {
                continue;
            }
            OpenApiGenerator.setEndpointResponseContentType(endpointOperation, responseContentType, statusCode, operationInfo.responses[statusCode]);
        }
    }
    static setEndpointResponseContentType(endpointOperation, responseContentType, statusCode, apiBodyInfo) {
        let response;
        if (apiBodyInfo) {
            // user defined response body info
            let responseContent = {};
            let mediaTypeObject = {};
            let responseType = (apiBodyInfo.contentType || responseContentType).toLowerCase();
            responseContent[responseType] = mediaTypeObject;
            response = {
                content: responseContent,
                description: apiBodyInfo.description || ""
            };
            if (apiBodyInfo.type) {
                OpenApiGenerator.addPrimitiveToMediaTypeObject(mediaTypeObject, apiBodyInfo);
            }
            else if (apiBodyInfo.class) {
                OpenApiGenerator.addClassToMediaTypeObject(mediaTypeObject, apiBodyInfo, responseType);
            }
            if (apiBodyInfo.example) {
                mediaTypeObject.example = apiBodyInfo.example;
            }
            else if (mediaTypeObject.schema) {
                let schema = mediaTypeObject.schema;
                if (schema && schema.example) {
                    mediaTypeObject.example = schema.example;
                }
            }
        }
        else {
            // response content type only
            let responseContent = {};
            responseContent[responseContentType] = {};
            response = {
                content: responseContent,
                description: "" // required or swagger ui will throw errors
            };
        }
        if (statusCode) {
            endpointOperation.responses[statusCode] = response;
        }
        else {
            endpointOperation.responses.default = response;
        }
    }
    static addPrimitiveToMediaTypeObject(mediaTypeObject, apiBodyInfo) {
        let type = apiBodyInfo.type.toLowerCase();
        if (!OpenApiGenerator.OPEN_API_TYPES.includes(type)) {
            return;
        }
        let schemaType = OpenApiGenerator.OPEN_API_SCHEMA_TYPE_MAP[apiBodyInfo.type];
        if (type !== "file") {
            mediaTypeObject.schema = {
                example: OpenApiGenerator.getPrimitiveTypeExample(apiBodyInfo.type),
                type: schemaType
            };
        }
        else {
            mediaTypeObject.schema = {
                format: "binary",
                type: schemaType
            };
        }
    }
    static getPrimitiveTypeExample(type) {
        let example = OpenApiGenerator.OPEN_API_TYPE_EXAMPLES[type];
        return jsonUtils_1.toJson(example);
    }
    static addClassToMediaTypeObject(mediaTypeObject, apiBodyInfo, responseContentType) {
        let clazz = apiBodyInfo.class;
        let instance;
        if ((typeof clazz.example) === "function") {
            instance = clazz.example();
        }
        else {
            instance = new clazz();
        }
        let schema = {
            properties: {},
            type: "object"
        };
        mediaTypeObject.schema = schema;
        if (responseContentType === "application/json") {
            let exampleJson = jsonUtils_1.toJson(instance);
            mediaTypeObject.example = exampleJson;
            schema.example = exampleJson;
        }
        OpenApiGenerator.addObjectPropertiesToSchema(mediaTypeObject.schema, instance);
    }
    static addObjectPropertiesToSchema(schema, instance) {
        let objectProperties = Object.getOwnPropertyNames(instance);
        for (let property of objectProperties) {
            let type = OpenApiGenerator.getTypeOfInstanceProperty(instance, property);
            if (!OpenApiGenerator.OPEN_API_TYPES.includes(type)) {
                // not a supported type or null/undefined, ommit this property
                continue;
            }
            let propertySchema = {
                type
            };
            if (type === "object") {
                // get schema for property object
                propertySchema.example = jsonUtils_1.toJson(instance[property]);
                OpenApiGenerator.addObjectPropertiesToSchema(propertySchema, instance[property]);
            }
            else if (type === "array") {
                if (instance[property].length > 0) {
                    if (!OpenApiGenerator.addArrayToSchema(propertySchema, instance[property])) {
                        // unsupported array item type or null/undefined, ommit this property
                        continue;
                    }
                }
                else {
                    // no way to determine array item type, so better to ommit this property
                    continue;
                }
            }
            else {
                propertySchema.example = instance[property];
            }
            if (!schema.properties) {
                schema.properties = {};
            }
            schema.properties[property] = propertySchema;
        }
    }
    static addArrayToSchema(propertySchema, instance) {
        let itemType = OpenApiGenerator.getInstanceType(instance[0]);
        if (!OpenApiGenerator.OPEN_API_TYPES.includes(itemType)) {
            return false;
        }
        propertySchema.example = jsonUtils_1.toJson(instance);
        propertySchema.items = {
            type: itemType
        };
        if (itemType === "object") {
            propertySchema.items.example = jsonUtils_1.toJson(instance[0]);
            // get schema for array item type
            OpenApiGenerator.addObjectPropertiesToSchema(propertySchema.items, instance[0]);
        }
        else if (itemType === "array") {
            if (instance.length > 0) {
                OpenApiGenerator.addArrayToSchema(propertySchema.items, instance[0]);
            }
        }
        else {
            propertySchema.items.example = instance[0];
        }
        return true;
    }
    static getTypeOfInstanceProperty(instance, property) {
        return OpenApiGenerator.getInstanceType(instance[property]);
    }
    static getInstanceType(instance) {
        let type = ((typeof instance)).toLowerCase();
        if (type === "object" && Array.isArray(instance)) {
            // the property is actually an array
            type = "array";
        }
        return type;
    }
    static addParametersToEndpoint(endpointOperation, endpointInfo) {
        endpointOperation.parameters = [];
        endpointInfo.parameterExtractors.forEach(p => {
            if (p.source === "virtual") {
                return;
            }
            let paramInfo = {
                in: p.source,
                name: p.name,
                schema: {} // required, or breaks query parameters
            };
            if (p.source === "path") {
                // swagger ui will break if this is not set to true
                paramInfo.required = true;
            }
            endpointOperation.parameters.push(paramInfo);
        });
    }
}
OpenApiGenerator.ENDPOINTS = DecoratorRegistry_1.DecoratorRegistry.Endpoints;
OpenApiGenerator.AUTH_FILTERS = DecoratorRegistry_1.DecoratorRegistry.AuthFilters;
OpenApiGenerator.OPEN_API_TYPES = [
    "array",
    "array-array",
    "boolean",
    "boolean-array",
    "double",
    "double-array",
    "file",
    "int",
    "int-array",
    "number",
    "number-array",
    "object",
    "object-array",
    "string",
    "string-array"
];
OpenApiGenerator.OPEN_API_SCHEMA_TYPE_MAP = {
    "array": "array",
    "array-array": "array",
    "boolean": "boolean",
    "boolean-array": "array",
    "double": "number",
    "double-array": "array",
    "file": "string",
    "int": "number",
    "int-array": "array",
    "number": "number",
    "number-array": "array",
    "object": "object",
    "object-array": "array",
    "string": "string",
    "string-array": "array"
};
OpenApiGenerator.OPEN_API_TYPE_EXAMPLES = {
    "array": [],
    "array-array": [[], [], []],
    "boolean": true,
    "boolean-array": [true, false, true],
    "double": 1.1,
    "double-array": [1.1, 2.2, 3.3],
    "file": "upload a file",
    "int": 1,
    "int-array": [1, 2, 3],
    "number": 1.1,
    "object": {},
    "object-array": [{}, {}, {}],
    "string": "a string",
    "string-array": ["1st string", "2nd string", "3rd string"]
};
__decorate([
    timed_1.timed,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [AppConfig_1.AppConfig, MiddlewareRegistry_1.MiddlewareRegistry]),
    __metadata("design:returntype", void 0)
], OpenApiGenerator, "buildOpenApiSpec", null);
__decorate([
    timed_1.timed,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, AppConfig_1.AppConfig,
        MiddlewareRegistry_1.MiddlewareRegistry]),
    __metadata("design:returntype", Promise)
], OpenApiGenerator, "exportOpenApiSpec", null);
exports.OpenApiGenerator = OpenApiGenerator;
//# sourceMappingURL=OpenApiGenerator.js.map